from pyzac import pyzac_decorator
from pyzac import started_processes
from pyzac import assertlist
from pyzac import debuglist
from pyzac import partial_sub
from pyzac import add_debug_info
from pyzac import create_pub_func
from pyzac import create_sub_socket
from multiprocessing import Process
import zmq
from time import sleep


def test_partial_sub():
    class mocksocket:
        def recv_pyobj(self, flags="notused"):
            return 20

    a = mocksocket()

    def atest(myparam, myparamtwo):
        return myparam * myparamtwo

    testval = partial_sub(atest, a)(1)
    assert testval == 20


def test_partial_sub_addition():
    counter = 20

    class mocksocket:
        def recv_pyobj(self, flags="notused"):
            nonlocal counter
            counter = counter + 1
            return counter

    a = mocksocket()

    def atest(myparam, myparamtwo):
        return myparam * myparamtwo

    myfunc = partial_sub(partial_sub(atest, a), a)
    testval = myfunc()
    assert testval == 21 * 22
    testval = myfunc()
    assert testval == 23 * 24
    testval = myfunc()
    assert testval == 25 * 26


def test_zmq():
    recfile = open("test.txt", "w")

    def pub_func():
        context = zmq.Context()
        #       socket = context.socket(zmq.PUB)
        #      socket.bind("tcp://127.0.0.1:2000")

        # Allow clients to connect before sending data
        #     sleep(1)
        #    socket.send_pyobj(20)

        def atest():
            sleep(1)
            return 20

        context = zmq.Context()
        newfunc = create_pub_func(context, atest, "tcp://127.0.0.1:2000")
        newfunc()

    def sub_func():
        context = zmq.Context()
        socket = context.socket(zmq.SUB)
        # We can connect to several endpoints if we desire, and receive from all.
        socket.connect("tcp://127.0.0.1:2000")

        # We must declare the socket as of type SUBSCRIBER, and pass a prefix filter.
        # Here, the filter is the empty string, wich means we receive all messages.
        # We may subscribe to several filters, thus receiving from all.
        socket.setsockopt(zmq.SUBSCRIBE, b"")

        message = socket.recv_pyobj()
        print(message)

    print("start_new_func")
    pub_process = Process(target=pub_func)
    pub_process.start()
    print("start_sub_func")
    rec_process = Process(target=sub_func)
    rec_process.start()
    sleep(2)
    pub_process.terminate()
    pub_process.join()
    rec_process.terminate()
    rec_process.join()


def test_partial_pub_zmq():
    def atest():
        return 20

    def run_newfunc():
        context = zmq.Context()
        newfunc = create_pub_func(context, atest, "tcp://127.0.0.1:5000")
        while True:
            assert 20 == newfunc()

    def run_receive():
        subcontext = zmq.Context()
        sub_socket = create_sub_socket("tcp://127.0.0.1:5000", subcontext)
        while True:
            testobj = sub_socket.recv_pyobj()
            print(testobj)
            assert testobj == 20

    pub_process = Process(target=run_newfunc)
    pub_process.start()
    sleep(1)
    rec_process = Process(target=run_receive)
    rec_process.start()
    sleep(1)
    pub_process.terminate()
    pub_process.join()
    rec_process.terminate()
    rec_process.join()


def test_partial_pub_sub_zmq():
    def atest():
        return 20

    def get_test(avalue):
        assert (avalue == 20) or (avalue == 0)

    def run_newfunc():
        context = zmq.Context()
        newfunc = create_pub_func(context, atest, "tcp://127.0.0.1:5000")
        while True:
            assert 20 == newfunc()

    def run_receive():
        subcontext = zmq.Context()
        sub_socket = create_sub_socket("tcp://127.0.0.1:5000", subcontext)

        recfunc = partial_sub(get_test, sub_socket)
        while True:
            recfunc()

    pub_process = Process(target=run_newfunc)
    pub_process.start()
    sleep(1)
    rec_process = Process(target=run_receive)
    rec_process.start()
    sleep(1)
    pub_process.terminate()
    pub_process.join()
    rec_process.terminate()
    rec_process.join()


# assert len(reclist) > 0


def test_partial_sub_addition_keyarg():
    """
    This function is used to test the keyargument
    handling in the partial_sub function.
    """
    startcounter = 20
    counter = startcounter

    class mocksocket:
        def recv_pyobj(self, flags="notused"):
            nonlocal counter
            counter = counter + 1
            return counter

    a = mocksocket()

    def atest(myparam, myparamtwo=0):
        return myparam * myparamtwo

    myfunc = partial_sub(partial_sub(atest, a), a, keyargname="myparamtwo")
    # the myfunc has to be called multiple times to simulate multiple receive operations
    # for the encapsulated sockets
    testval = myfunc()
    assert testval == (startcounter + 1) * (startcounter + 2)
    testval = myfunc()
    assert testval == (startcounter + 3) * (startcounter + 4)
    testval = myfunc()
    assert testval == (startcounter + 5) * (startcounter + 6)


def test_decorators():
    @pyzac_decorator(pub_addr="tcp://localhost:2000")
    def publisher():
        add_debug_info("in publisher")
        return 20

    @pyzac_decorator(pos_sub_addr=["tcp://localhost:2000"])
    def subscriber(result):
        add_debug_info("in subscriber")
        print(result)
        # assert result == 20

    publisher()
    sleep(1)
    subscriber()
    sleep(1)

    # assert len(debuglist) != 0
    # for dinfo in debuglist:
    #     print(dinfo)

    # assert len(assertlist) != 0
    # for dinfo in assertlist:
    #    print(dinfo)

    for p in started_processes:
        p.terminate()
        p.join()


#
# def test_paramsubmap_decorators():
#     @pyzac_decorator(pub_addr="tcp://127.0.0.1:2000")
#     def publisher():
#         # print("hello")
#         return 20
#
#     @pyzac_decorator(pos_sub_addr=["tcp://localhost:2000"])
#     def subscriber(result):
#         print(result)
#         assert result == 20
#
#     publisher()
#     sleep(1)
#     subscriber()
#
#     assert len(assertlist) != 0
#     for dinfo in assertlist:
#         print(dinfo)
#
#     for dinfo in debuglist:
#         print(dinfo)
#
#     for p in started_processes:
#         p.terminate()
#         p.join()

#
# def test_decorator_mesh():
#     @pyzac_decorator(pub_addr="tcp://127.0.0.1:2000")
#     def publisher():
#         return 20
#
#     @pyzac_decorator(
#         pos_sub_addr="tcp://localhost:2000", pub_addr="tcp://127.0.0.1:2001"
#     )
#     def filter(result):
#         assert result == 30
#         return 40
#
#     @pyzac_decorator(pos_sub_addr="tcp://localhost:2001")
#     def end_point(result):
#         assert result == 40
#
#     publisher()
#     sleep(1)
#     filter()
#     sleep(1)
#     end_point()
#     sleep(1)
#
#     assert len(assertlist) != 0
#     for dinfo in assertlist:
#         print(dinfo)
#
#     for p in started_processes:
#         p.terminate()
#         p.join()


def test_state():
    count = 0
    std_pub_val = 20

    @pyzac_decorator(pub_addr="tcp://127.0.0.1:2000")
    def publisher():
        return std_pub_val

    @pyzac_decorator(pos_sub_addr="tcp://localhost:2000")
    def subscriber(result, pyzac_state=0):
        nonlocal count
        if count == 1:
            assert pyzac_state > 0
            assert pyzac_state == std_pub_val
        if count == 2:
            assert pyzac_state == 2 * std_pub_val
        count = count + 1
        print(count)
        return result + pyzac_state

    publisher()
    sub = subscriber
    sub()
    sleep(1)
    for p in started_processes:
        p.terminate()
        p.join()
        print("Processes stoped")


if __name__ == "__main__":
    test_decorator_mesh()
    test_state()



import zmq
from multiprocessing import Process
import functools
import inspect
import sys

started_processes = list()

debuglist = list()
assertlist = list()

cstatekey = "pyzac_state"


lrsocket = {}

c_def_argvalue = 1
c_def_blank = ""


def excepthook(*args):
    assertlist.append(args)


sys.excepthook = excepthook


def add_debug_info(debugmessage):
    debuglist.append(debugmessage)


def create_sub_socket(sub_addr, cntx):
    sock_sub = cntx.socket(zmq.SUB)
    sock_sub.connect(sub_addr)
    sock_sub.setsockopt_string(zmq.SUBSCRIBE, "")
    # sock_sub.setsockopt(zmq.SUBSCRIBE, b"")
    return sock_sub


def extract_func_parameters(input_sockets):
    for param in input_sockets:
        try:
            input_sockets[param].recv_pyobj(flags=zmq.NOBLOCK)
            # check for a message, this will not block
        except zmq.Again as e:
            pass


def _try_receive_arg_from_socket(sub_socket):
    try:
        argvalue = sub_socket.recv_pyobj(flags=zmq.NOBLOCK)
        # check for a message, this will not block
    except zmq.Again as e:
        try:
            argvalue = lrsocket[sub_socket]
        except:
            pass
        pass
    lrsocket[sub_socket] = argvalue
    return argvalue


def partial_sub(func, sub_socket, keyargname=c_def_blank, def_arg_value=c_def_argvalue):
    lrsocket[sub_socket] = def_arg_value

    def resultfunc(*fargs, **fkeywords):
        argvalue = _try_receive_arg_from_socket(sub_socket)
        add_debug_info(argvalue)
        if keyargname != c_def_blank:
            newkeywords = {keyargname: argvalue}
            newkeywords.update(fkeywords)
            return func(*fargs, **newkeywords)
        else:
            return func(argvalue, *fargs, **fkeywords)

    resultfunc.func = func
    resultfunc.sub_socket = sub_socket
    return resultfunc


def _pub_wrapper(func, pub_socket):
    """
    Pub_wrapper is used to add the zero_mq publish part to the function.
    :param func:
    :param pub_socket: socket which is used for mapping values to parameters
    :param param_name: name of the parameter which is mapped onto the socket
    :return:
    """

    def newfunc(*fargs, **fkeywords):
        func_res = None
        try:
            func_res = func(*fargs, **fkeywords)
            # print(pub_socket.get(zmq.ZMQ_LAST_ENDPOINT))
            if func_res != None:
                pub_socket.send_pyobj(func_res)
        except:
            print("exception")
            Exception("Cannot send pub")
            pass
        return func_res

    newfunc.func = func
    newfunc.pub_socket = pub_socket
    return newfunc


def get_pos_and_key_names(func):
    params = inspect.signature(func).parameters
    pos_arg_names = [k for k in params if params[k].default == inspect._empty]
    key_arg_names = [k for k in params if params[k].default != inspect._empty]
    return pos_arg_names, key_arg_names


def _wrap_pyzmq(func, pub_addr="", pos_sub_addr=[], key_sub_addr={}, pyzac_state={}):
    """
    :param func:
    :param pub_addr: all generated results are distributed to that address
    :param sub_addr: dictionary mapping parameters to sockets, key is parameter name value equal to address
    :return:

    """
    context = zmq.Context()

    posnames, keynames = get_pos_and_key_names(func)

    new_key_sub_addr = _check_mapping_of_args(
        key_sub_addr, keynames, pos_sub_addr, posnames, pyzac_state
    )
    in_sockets = _create_socket_mapping(context, new_key_sub_addr)
    # pos_sub_addr, posnames)

    pub = not (pub_addr == "")

    newfunc = _create_partial_func(func, in_sockets, new_key_sub_addr)  # , posnames)

    if pub:
        newfunc = create_pub_func(context, newfunc, pub_addr)

    if len(pyzac_state.keys()) == 0:
        while True:
            newfunc()
    else:
        while True:
            value = newfunc(pyzac_state=pyzac_state)
            if not isinstance(value, list):
                value = [value]

            # pyzac_state = dict(zip(pyzac_state.keys(), value))


def create_pub_func(context, newfunc, pub_addr):
    sock_pub = context.socket(zmq.PUB)
    sock_pub.bind(pub_addr)
    newfunc = _pub_wrapper(newfunc, sock_pub)
    return newfunc


def _create_partial_func(func, in_sockets, key_sub_addr):  # , posnames):
    newfunc = func
    # for sub in posnames:
    #     newfunc = partial_sub(newfunc, in_sockets[sub])
    for key in key_sub_addr:
        newfunc = partial_sub(newfunc, in_sockets[key], keyargname=key)
    return newfunc


def _create_socket_mapping(context, key_sub_addr):  # , pos_sub_addr, posnames):
    in_sockets = {}
    counter = 0
    # for posname in posnames:
    #     in_sockets[posname] = create_sub_socket(pos_sub_addr[counter], context)
    #     counter += 1
    for sub in key_sub_addr:
        in_sockets[sub] = create_sub_socket(key_sub_addr[sub], context)
    return in_sockets


def _check_mapping_of_args(key_sub_addr, keynames, pos_sub_addr, posnames, state=None):
    """
    :param key_sub_addr: dictionary key= name of keyword argument value= address of zmq publisher
    :param keynames: list of keyword arguments
    :param pos_sub_addr: list of addresses to zmq publishers for the positional arguments
    :param posnames: positional arguments
    :return: returns new composed dictionary containing positional and keyword arguments mappings
    """
    num_state = 0
    if isinstance(state, dict):
        num_state = len(state.keys())

    key_length = len(keynames)
    pos_length = len(posnames)
    addr_length = len(pos_sub_addr) + len(key_sub_addr.keys()) + num_state

    not_key_args_mapped = not len(key_sub_addr.keys()) == key_length
    not_pos_args_mapped = not len(posnames) == pos_length
    map_length_correct = addr_length == (key_length + pos_length)

    result = dict(**key_sub_addr, **dict(zip(posnames, pos_sub_addr)))
    # Delete state keys from target key dict
    for k in state.keys():
        result.pop(k, None)

    if map_length_correct:
        return result

    if not_key_args_mapped or not_pos_args_mapped:
        if not_key_args_mapped:
            raise Exception("key args not mapped")
        if not_pos_args_mapped:
            raise Exception("pos args not mapped")


def pyzac_decorator(
    pub_addr="", pos_sub_addr=None, key_sub_addr=None, pyzac_state=None
):
    key_addr = key_sub_addr
    sub_addr = pos_sub_addr
    state = pyzac_state
    if key_sub_addr == None:
        key_addr = {}

    if pos_sub_addr == None:
        sub_addr = []

    if pyzac_state == None:
        state = {}

    def decorator_pyzeromq(func):
        @functools.wraps(func)
        def wrapper_process(*args, **kwargs):
            # partial is used to generate a function with no parameters
            # from _wrap_pyzmq and the input parameters are fixed to
            # func pub_addr and sub_addr
            f = functools.partial(
                _wrap_pyzmq, func, pub_addr, sub_addr, key_addr, state
            )
            new_process = Process(target=f)
            new_process.start()
            # next line is used to track the started processes
            started_processes.append(new_process)

        return wrapper_process

    return decorator_pyzeromq



#!/bin/env python

"""
Setup script for rinohtype
"""

import os
import sys

from setuptools import setup, find_packages


BASE_PATH = os.path.dirname(os.path.abspath(__file__))


if sys.version_info < (3, 3):
    print("pyzac requires Python 3.3 or higher")
    sys.exit(1)


def long_description():
    with open(os.path.join(BASE_PATH, "README.rst")) as readme:
        result = readme.read()
    result += "\n\n"
    with open(os.path.join(BASE_PATH, "CHANGES.rst")) as changes:
        result += changes.read()
    return result


setup(
    name="pyzac",
    version="0.1.4",
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    python_requires=">= 3.3",
    license="Apache 2.0",
    # install_requires=["setuptools", "pip", "docutils", "purepng>=0.1.1"],
    # extras_require={"bitmap": ["Pillow"]},
    # setup_requires=["pytest-runner"],
    # tests_require=["pytest>=2.0.0", "pytest-assume", "requests", "PyPDF2"],
    author="Dark Ligt alias FB2011B",
    description="The Python Actor Model using zeroMQ",
    long_description=open("README.rst").read(),
    url="https://github.com/F2011B/pyzac",
    keywords="actor zeromq pzmq",
    classifiers=[
        "Environment :: Console",
        "Environment :: Other Environment",
        "Environment :: Web Environment",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
